print("import: 'shap'")
import shap

print("import: 'shap.explainers'")
import shap.explainers

print("import: 'shap.explainers.other'")
import shap.explainers.other

print("import: 'shap.explainers._deep'")
import shap.explainers._deep

print("import: 'shap.plots'")
import shap.plots

print("import: 'shap.plots.colors'")
import shap.plots.colors

print("import: 'shap.benchmark'")
import shap.benchmark

print("import: 'shap.maskers'")
import shap.maskers

print("import: 'shap.utils'")
import shap.utils

print("import: 'shap.actions'")
import shap.actions

print("import: 'shap.models'")
import shap.models

